import branca.colormap as colormap
import branca.element as element

try:
    from ._version import __version__
except ImportError:
    __version__ = "unknown"


__all__ = [
    "colormap",
    "element",
]



"""
Colormap
--------

Utility module for dealing with colormaps.

"""

import json
import math
import os
from typing import Dict, List, Optional, Sequence, Tuple, Union

from jinja2 import Template

from branca.element import ENV, Figure, JavascriptLink, MacroElement
from branca.utilities import legend_scaler

rootpath: str = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(rootpath, "_cnames.json")) as f:
    _cnames: Dict[str, str] = json.loads(f.read())

with open(os.path.join(rootpath, "_schemes.json")) as f:
    _schemes: Dict[str, List[str]] = json.loads(f.read())


TypeRGBInts = Tuple[int, int, int]
TypeRGBFloats = Tuple[float, float, float]
TypeRGBAInts = Tuple[int, int, int, int]
TypeRGBAFloats = Tuple[float, float, float, float]
TypeAnyColorType = Union[TypeRGBInts, TypeRGBFloats, TypeRGBAInts, TypeRGBAFloats, str]


def _is_hex(x: str) -> bool:
    return x.startswith("#") and len(x) == 7


def _parse_hex(color_code: str) -> TypeRGBAFloats:
    return (
        _color_int_to_float(int(color_code[1:3], 16)),
        _color_int_to_float(int(color_code[3:5], 16)),
        _color_int_to_float(int(color_code[5:7], 16)),
        1.0,
    )


def _color_int_to_float(x: int) -> float:
    """Convert an integer between 0 and 255 to a float between 0. and 1.0"""
    return x / 255.0


def _color_float_to_int(x: float) -> int:
    """Convert a float between 0. and 1.0 to an integer between 0 and 255"""
    return int(x * 255.9999)


def _parse_color(x: Union[tuple, list, str]) -> TypeRGBAFloats:
    if isinstance(x, (tuple, list)):
        return tuple(tuple(x) + (1.0,))[:4]  # type: ignore
    elif isinstance(x, str) and _is_hex(x):
        return _parse_hex(x)
    elif isinstance(x, str):
        cname = _cnames.get(x.lower(), None)
        if cname is None:
            raise ValueError(f"Unknown color {cname!r}.")
        return _parse_hex(cname)
    else:
        raise ValueError(f"Unrecognized color code {x!r}")


def _base(x: float) -> float:
    if x > 0:
        base = pow(10, math.floor(math.log10(x)))
        return round(x / base) * base
    else:
        return 0


class ColorMap(MacroElement):
    """A generic class for creating colormaps.

    Parameters
    ----------
    vmin: float
        The left bound of the color scale.
    vmax: float
        The right bound of the color scale.
    caption: str
        A caption to draw with the colormap.
    text_color: str, default "black"
        The color for the text.
    max_labels : int, default 10
        Maximum number of legend tick labels
    """

    _template: Template = ENV.get_template("color_scale.js")

    def __init__(
        self,
        vmin: float = 0.0,
        vmax: float = 1.0,
        caption: str = "",
        text_color: str = "black",
        max_labels: int = 10,
    ):
        super().__init__()
        self._name = "ColorMap"

        self.vmin = vmin
        self.vmax = vmax
        self.caption = caption
        self.text_color = text_color
        self.index: List[float] = [vmin, vmax]
        self.max_labels = max_labels
        self.tick_labels: Optional[Sequence[Union[float, str]]] = None

        self.width = 450
        self.height = 40

    def render(self, **kwargs):
        """Renders the HTML representation of the element."""
        self.color_domain = [
            float(self.vmin + (self.vmax - self.vmin) * k / 499.0) for k in range(500)
        ]
        self.color_range = [self.__call__(x) for x in self.color_domain]

        # sanitize possible numpy floats to native python floats
        self.index = [float(i) for i in self.index]

        if self.tick_labels is None:
            self.tick_labels = legend_scaler(self.index, self.max_labels)

        super().render(**kwargs)

        figure = self.get_root()
        assert isinstance(figure, Figure), (
            "You cannot render this Element " "if it is not in a Figure."
        )

        figure.header.add_child(
            JavascriptLink("https://cdnjs.cloudflare.com/ajax/libs/d3/3.5.5/d3.min.js"),
            name="d3",
        )  # noqa

    def rgba_floats_tuple(self, x: float) -> TypeRGBAFloats:
        """
        This class has to be implemented for each class inheriting from
        Colormap. This has to be a function of the form float ->
        (float, float, float, float) describing for each input float x,
        the output color in RGBA format;
        Each output value being between 0 and 1.
        """
        raise NotImplementedError

    def rgba_bytes_tuple(self, x: float) -> TypeRGBAInts:
        """Provides the color corresponding to value `x` in the
        form of a tuple (R,G,B,A) with int values between 0 and 255.
        """
        return tuple(_color_float_to_int(u) for u in self.rgba_floats_tuple(x))  # type: ignore

    def rgb_bytes_tuple(self, x: float) -> TypeRGBInts:
        """Provides the color corresponding to value `x` in the
        form of a tuple (R,G,B) with int values between 0 and 255.
        """
        return self.rgba_bytes_tuple(x)[:3]

    def rgb_hex_str(self, x: float) -> str:
        """Provides the color corresponding to value `x` in the
        form of a string of hexadecimal values "#RRGGBB".
        """
        return "#%02x%02x%02x" % self.rgb_bytes_tuple(x)

    def rgba_hex_str(self, x: float) -> str:
        """Provides the color corresponding to value `x` in the
        form of a string of hexadecimal values "#RRGGBBAA".
        """
        return "#%02x%02x%02x%02x" % self.rgba_bytes_tuple(x)

    def __call__(self, x: float) -> str:
        """Provides the color corresponding to value `x` in the
        form of a string of hexadecimal values "#RRGGBBAA".
        """
        return self.rgba_hex_str(x)

    def _repr_html_(self) -> str:
        """Display the colormap in a Jupyter Notebook.

        Does not support all the class arguments.

        """
        nb_ticks = 7
        delta_x = math.floor(self.width / (nb_ticks - 1))
        x_ticks = [(i) * delta_x for i in range(0, nb_ticks)]
        delta_val = delta_x * (self.vmax - self.vmin) / self.width
        val_ticks = [round(self.vmin + (i) * delta_val, 1) for i in range(0, nb_ticks)]

        return (
            f'<svg height="40" width="{self.width}">'
            + "".join(
                [
                    (
                        '<line x1="{i}" y1="15" x2="{i}" '
                        'y2="27" style="stroke:{color};stroke-width:2;" />'
                    ).format(
                        i=i * 1,
                        color=self.rgba_hex_str(
                            self.vmin + (self.vmax - self.vmin) * i / (self.width - 1),
                        ),
                    )
                    for i in range(self.width)
                ],
            )
            + (
                '<text x="0" y="38" style="text-anchor:start; font-size:11px;'
                ' font:Arial; fill:{}">{}</text>'
            ).format(
                self.text_color,
                self.vmin,
            )
            + "".join(
                [
                    (
                        '<text x="{}" y="38"; style="text-anchor:middle; font-size:11px;'
                        ' font:Arial; fill:{}">{}</text>'
                    ).format(x_ticks[i], self.text_color, val_ticks[i])
                    for i in range(1, nb_ticks - 1)
                ],
            )
            + (
                '<text x="{}" y="38" style="text-anchor:end; font-size:11px;'
                ' font:Arial; fill:{}">{}</text>'
            ).format(
                self.width,
                self.text_color,
                self.vmax,
            )
            + '<text x="0" y="12" style="font-size:11px; font:Arial; fill:{}">{}</text>'.format(
                self.text_color,
                self.caption,
            )
            + "</svg>"
        )


class LinearColormap(ColorMap):
    """Creates a ColorMap based on linear interpolation of a set of colors
    over a given index.

    Parameters
    ----------

    colors : list-like object with at least two colors.
        The set of colors to be used for interpolation.
        Colors can be provided in the form:
        * tuples of RGBA ints between 0 and 255 (e.g: `(255, 255, 0)` or
        `(255, 255, 0, 255)`)
        * tuples of RGBA floats between 0. and 1. (e.g: `(1.,1.,0.)` or
        `(1., 1., 0., 1.)`)
        * HTML-like string (e.g: `"#ffff00`)
        * a color name or shortcut (e.g: `"y"` or `"yellow"`)
    index : list of floats, default None
        The values corresponding to each color.
        It has to be sorted, and have the same length as `colors`.
        If None, a regular grid between `vmin` and `vmax` is created.
    vmin : float, default 0.
        The minimal value for the colormap.
        Values lower than `vmin` will be bound directly to `colors[0]`.
    vmax : float, default 1.
        The maximal value for the colormap.
        Values higher than `vmax` will be bound directly to `colors[-1]`.
    caption: str
        A caption to draw with the colormap.
    text_color: str, default "black"
        The color for the text.
    max_labels : int, default 10
        Maximum number of legend tick labels
    tick_labels: list of floats, default None
        If given, used as the positions of ticks."""

    def __init__(
        self,
        colors: Sequence[TypeAnyColorType],
        index: Optional[Sequence[float]] = None,
        vmin: float = 0.0,
        vmax: float = 1.0,
        caption: str = "",
        text_color: str = "black",
        max_labels: int = 10,
        tick_labels: Optional[Sequence[float]] = None,
    ):
        super().__init__(
            vmin=vmin,
            vmax=vmax,
            caption=caption,
            text_color=text_color,
            max_labels=max_labels,
        )
        self.tick_labels: Optional[Sequence[float]] = tick_labels

        n = len(colors)
        if n < 2:
            raise ValueError("You must provide at least 2 colors.")
        if index is None:
            self.index = [vmin + (vmax - vmin) * i * 1.0 / (n - 1) for i in range(n)]
        else:
            self.index = list(index)
        self.colors: List[TypeRGBAFloats] = [_parse_color(x) for x in colors]

    def rgba_floats_tuple(self, x: float) -> TypeRGBAFloats:
        """Provides the color corresponding to value `x` in the
        form of a tuple (R,G,B,A) with float values between 0. and 1.
        """
        if x <= self.index[0]:
            return self.colors[0]
        if x >= self.index[-1]:
            return self.colors[-1]

        i = len([u for u in self.index if u < x])  # 0 < i < n.
        if self.index[i - 1] < self.index[i]:
            p = (x - self.index[i - 1]) * 1.0 / (self.index[i] - self.index[i - 1])
        elif self.index[i - 1] == self.index[i]:
            p = 1.0
        else:
            raise ValueError("Thresholds are not sorted.")

        return tuple(  # type: ignore
            (1.0 - p) * self.colors[i - 1][j] + p * self.colors[i][j] for j in range(4)
        )

    def to_step(
        self,
        n: Optional[int] = None,
        index: Optional[Sequence[float]] = None,
        data: Optional[Sequence[float]] = None,
        method: str = "linear",
        quantiles: Optional[Sequence[float]] = None,
        round_method: Optional[str] = None,
        max_labels: int = 10,
    ) -> "StepColormap":
        """Splits the LinearColormap into a StepColormap.

        Parameters
        ----------
        n : int, default None
            The number of expected colors in the output StepColormap.
            This will be ignored if `index` is provided.
        index : list of floats, default None
            The values corresponding to each color bounds.
            It has to be sorted.
            If None, a regular grid between `vmin` and `vmax` is created.
        data : list of floats, default None
            A sample of data to adapt the color map to.
        method : str, default 'linear'
            The method used to create data-based colormap.
            It can be 'linear' for linear scale, 'log' for logarithmic,
            or 'quant' for data's quantile-based scale.
        quantiles : list of floats, default None
            Alternatively, you can provide explicitly the quantiles you
            want to use in the scale.
        round_method : str, default None
            The method used to round thresholds.
            * If 'int', all values will be rounded to the nearest integer.
            * If 'log10', all values will be rounded to the nearest
            order-of-magnitude integer. For example, 2100 is rounded to
            2000, 2790 to 3000.
        max_labels : int, default 10
            Maximum number of legend tick labels

        Returns
        -------
        A StepColormap with `n=len(index)-1` colors.

        Examples:
        >> lc.to_step(n=12)
        >> lc.to_step(index=[0, 2, 4, 6, 8, 10])
        >> lc.to_step(data=some_list, n=12)
        >> lc.to_step(data=some_list, n=12, method='linear')
        >> lc.to_step(data=some_list, n=12, method='log')
        >> lc.to_step(data=some_list, n=12, method='quantiles')
        >> lc.to_step(data=some_list, quantiles=[0, 0.3, 0.7, 1])
        >> lc.to_step(data=some_list, quantiles=[0, 0.3, 0.7, 1],
        ...           round_method='log10')

        """
        msg = "You must specify either `index` or `n`"
        if index is None:
            if data is None:
                if n is None:
                    raise ValueError(msg)
                else:
                    index = [
                        self.vmin + (self.vmax - self.vmin) * i * 1.0 / n
                        for i in range(1 + n)
                    ]
                    scaled_cm = self
            else:
                max_ = max(data)
                min_ = min(data)
                scaled_cm = self.scale(vmin=min_, vmax=max_)
                method = "quantiles" if quantiles is not None else method
                if method.lower().startswith("lin"):
                    if n is None:
                        raise ValueError(msg)
                    index = [min_ + i * (max_ - min_) * 1.0 / n for i in range(1 + n)]
                elif method.lower().startswith("log"):
                    if n is None:
                        raise ValueError(msg)
                    if min_ <= 0:
                        msg = "Log-scale works only with strictly " "positive values."
                        raise ValueError(msg)
                    index = [
                        math.exp(
                            math.log(min_)
                            + i * (math.log(max_) - math.log(min_)) * 1.0 / n,
                        )
                        for i in range(1 + n)
                    ]
                elif method.lower().startswith("quant"):
                    if quantiles is None:
                        if n is None:
                            msg = (
                                "You must specify either `index`, `n` or" "`quantiles`."
                            )
                            raise ValueError(msg)
                        else:
                            quantiles = [i * 1.0 / n for i in range(1 + n)]
                    p = len(data) - 1
                    s = sorted(data)
                    index = [
                        s[int(q * p)] * (1.0 - (q * p) % 1)
                        + s[min(int(q * p) + 1, p)] * ((q * p) % 1)
                        for q in quantiles
                    ]
                else:
                    raise ValueError(f"Unknown method {method}")
        else:
            scaled_cm = self.scale(vmin=min(index), vmax=max(index))

        n = len(index) - 1

        if round_method == "int":
            index = [round(x) for x in index]

        if round_method == "log10":
            index = [_base(x) for x in index]

        colors = [
            scaled_cm.rgba_floats_tuple(
                index[i] * (1.0 - i / (n - 1.0)) + index[i + 1] * i / (n - 1.0),
            )
            for i in range(n)
        ]

        caption = self.caption
        text_color = self.text_color

        return StepColormap(
            colors,
            index=index,
            vmin=index[0],
            vmax=index[-1],
            caption=caption,
            text_color=text_color,
            max_labels=max_labels,
            tick_labels=self.tick_labels,
        )

    def scale(
        self,
        vmin: float = 0.0,
        vmax: float = 1.0,
        max_labels: int = 10,
    ) -> "LinearColormap":
        """Transforms the colorscale so that the minimal and maximal values
        fit the given parameters.
        """
        return LinearColormap(
            self.colors,
            index=[
                vmin + (vmax - vmin) * (x - self.vmin) * 1.0 / (self.vmax - self.vmin)
                for x in self.index
            ],  # noqa
            vmin=vmin,
            vmax=vmax,
            caption=self.caption,
            text_color=self.text_color,
            max_labels=max_labels,
        )


class StepColormap(ColorMap):
    """Creates a ColorMap based on linear interpolation of a set of colors
    over a given index.

    Parameters
    ----------
    colors : list-like object
        The set of colors to be used for interpolation.
        Colors can be provided in the form:
        * tuples of int between 0 and 255 (e.g: `(255,255,0)` or
        `(255, 255, 0, 255)`)
        * tuples of floats between 0. and 1. (e.g: `(1.,1.,0.)` or
        `(1., 1., 0., 1.)`)
        * HTML-like string (e.g: `"#ffff00`)
        * a color name or shortcut (e.g: `"y"` or `"yellow"`)
    index : list of floats, default None
        The bounds of the colors. The lower value is inclusive,
        the upper value is exclusive.
        It has to be sorted, and have the same length as `colors`.
        If None, a regular grid between `vmin` and `vmax` is created.
    vmin : float, default 0.
        The minimal value for the colormap.
        Values lower than `vmin` will be bound directly to `colors[0]`.
    vmax : float, default 1.
        The maximal value for the colormap.
        Values higher than `vmax` will be bound directly to `colors[-1]`.
    caption: str
        A caption to draw with the colormap.
    text_color: str, default "black"
        The color for the text.
    max_labels : int, default 10
        Maximum number of legend tick labels
    tick_labels: list of floats, default None
        If given, used as the positions of ticks.
    """

    def __init__(
        self,
        colors: Sequence[TypeAnyColorType],
        index: Optional[Sequence[float]] = None,
        vmin: float = 0.0,
        vmax: float = 1.0,
        caption: str = "",
        text_color: str = "black",
        max_labels: int = 10,
        tick_labels: Optional[Sequence[float]] = None,
    ):
        super().__init__(
            vmin=vmin,
            vmax=vmax,
            caption=caption,
            text_color=text_color,
            max_labels=max_labels,
        )
        self.tick_labels = tick_labels

        n = len(colors)
        if n < 1:
            raise ValueError("You must provide at least 1 colors.")
        if index is None:
            self.index = [vmin + (vmax - vmin) * i * 1.0 / n for i in range(n + 1)]
        else:
            self.index = list(index)
        self.colors: List[TypeRGBAFloats] = [_parse_color(x) for x in colors]

    def rgba_floats_tuple(self, x: float) -> TypeRGBAFloats:
        """
        Provides the color corresponding to value `x` in the
        form of a tuple (R,G,B,A) with float values between 0. and 1.

        """
        if x <= self.index[0]:
            return self.colors[0]
        if x >= self.index[-1]:
            return self.colors[-1]

        i = len([u for u in self.index if u <= x])  # 0 < i < n.
        return self.colors[i - 1]

    def to_linear(
        self,
        index: Optional[Sequence[float]] = None,
        max_labels: int = 10,
    ) -> LinearColormap:
        """
        Transforms the StepColormap into a LinearColormap.

        Parameters
        ----------
        index : list of floats, default None
                The values corresponding to each color in the output colormap.
                It has to be sorted.
                If None, a regular grid between `vmin` and `vmax` is created.
        max_labels : int, default 10
            Maximum number of legend tick labels

        """
        if index is None:
            n = len(self.index) - 1
            index = [
                self.index[i] * (1.0 - i / (n - 1.0))
                + self.index[i + 1] * i / (n - 1.0)
                for i in range(n)
            ]

        colors = [self.rgba_floats_tuple(x) for x in index]
        return LinearColormap(
            colors,
            index=index,
            vmin=self.vmin,
            vmax=self.vmax,
            caption=self.caption,
            text_color=self.text_color,
            max_labels=max_labels,
        )

    def scale(
        self,
        vmin: float = 0.0,
        vmax: float = 1.0,
        max_labels: int = 10,
    ) -> "StepColormap":
        """Transforms the colorscale so that the minimal and maximal values
        fit the given parameters.
        """
        return StepColormap(
            self.colors,
            index=[
                vmin + (vmax - vmin) * (x - self.vmin) * 1.0 / (self.vmax - self.vmin)
                for x in self.index
            ],  # noqa
            vmin=vmin,
            vmax=vmax,
            caption=self.caption,
            text_color=self.text_color,
            max_labels=max_labels,
        )


class _LinearColormaps:
    """A class for hosting the list of built-in linear colormaps."""

    def __init__(self):
        self._schemes = _schemes.copy()
        self._colormaps = {key: LinearColormap(val) for key, val in _schemes.items()}
        for key, val in _schemes.items():
            setattr(self, key, LinearColormap(val))

    def _repr_html_(self) -> str:
        return Template(
            """
        <table>
        {% for key,val in this._colormaps.items() %}
        <tr><td>{{key}}</td><td>{{val._repr_html_()}}</td></tr>
        {% endfor %}</table>
        """,
        ).render(this=self)


linear = _LinearColormaps()


class _StepColormaps:
    """A class for hosting the list of built-in step colormaps."""

    def __init__(self):
        self._schemes = _schemes.copy()
        self._colormaps = {key: StepColormap(val) for key, val in _schemes.items()}
        for key, val in _schemes.items():
            setattr(self, key, StepColormap(val))

    def _repr_html_(self) -> str:
        return Template(
            """
        <table>
        {% for key,val in this._colormaps.items() %}
        <tr><td>{{key}}</td><td>{{val._repr_html_()}}</td></tr>
        {% endfor %}</table>
        """,
        ).render(this=self)


step = _StepColormaps()



""""
Folium Colormap Module
----------------------
"""

import pytest

import branca.colormap as cm


def test_simple_step():
    step = cm.StepColormap(
        ["green", "yellow", "red"],
        vmin=3.0,
        vmax=10.0,
        index=[3, 4, 8, 10],
        caption="step",
    )
    step = cm.StepColormap(["r", "y", "g", "c", "b", "m"])
    step._repr_html_()


def test_simple_linear():
    linear = cm.LinearColormap(["green", "yellow", "red"], vmin=3.0, vmax=10.0)
    linear = cm.LinearColormap(
        ["red", "orange", "yellow", "green"],
        index=[0, 0.1, 0.9, 1.0],
    )
    linear._repr_html_()


black = "#000000ff"
red = "#ff0000ff"
green = "#00ff00ff"
blue = "#0000ffff"


def test_step_color_indexing():
    step = cm.StepColormap(colors=["black", "red", "lime", "blue"], index=[1, 2, 4, 5])
    assert step(0.99) == black
    assert step(1) == black
    assert step(1.01) == black
    assert step(1.99) == black
    assert step(2) == red
    assert step(2.01) == red
    assert step(3.99) == red
    assert step(4) == green
    assert step(4.01) == green
    assert step(4.99) == green
    assert step(5) == blue
    assert step(5.01) == blue


def test_step_color_indexing_larger_index():
    # add an upper bound to the last color, which doesn't do much but shouldn't fail
    step = cm.StepColormap(
        colors=["black", "red", "lime", "blue"],
        index=[1, 2, 4, 5, 10],
    )
    assert step(4.99) == green
    assert step(5) == blue
    assert step(10) == blue
    assert step(20) == blue


def test_linear_color_indexing():
    linear = cm.LinearColormap(
        colors=["black", "red", "lime", "blue"],
        index=[1, 2, 4, 5],
    )
    assert linear(1) == black
    assert linear(2) == red
    assert linear(4) == green
    assert linear(5) == blue
    assert linear(3) == "#7f7f00ff"


def test_linear_to_step():
    some_list = [30.6, 50, 51, 52, 53, 54, 55, 60, 70, 100]
    lc = cm.linear.YlOrRd_06
    lc.to_step(n=12)
    lc.to_step(index=[0, 2, 4, 6, 8, 10])
    lc.to_step(data=some_list, n=12)
    lc.to_step(data=some_list, n=12, method="linear")
    lc.to_step(data=some_list, n=12, method="log")
    lc.to_step(data=some_list, n=30, method="quantiles")
    lc.to_step(data=some_list, quantiles=[0, 0.3, 0.7, 1])
    lc.to_step(data=some_list, quantiles=[0, 0.3, 0.7, 1], round_method="int")
    lc.to_step(data=some_list, quantiles=[0, 0.3, 0.7, 1], round_method="log10")


def test_step_to_linear():
    step = cm.StepColormap(
        ["green", "yellow", "red"],
        vmin=3.0,
        vmax=10.0,
        index=[3, 4, 8, 10],
        caption="step",
    )
    step.to_linear()


def test_linear_object():
    cm.linear.OrRd_06._repr_html_()
    cm.linear.PuBu_06.to_step(12)
    cm.linear.YlGn_06.scale(3, 12)
    cm.linear._repr_html_()


def test_step_object():
    cm.step.OrRd_06._repr_html_()
    cm.step.PuBu_06.to_linear()
    cm.step.YlGn_06.scale(3, 12)
    cm.step._repr_html_()


@pytest.mark.parametrize(
    "max_labels,expected",
    [
        (10, [0.0, 1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]),
        (5, [0.0, "", 2.0, "", 4.0, "", 6.0, "", 8.0, ""]),
        (3, [0.0, "", "", "", 4.0, "", "", "", 8.0, "", "", ""]),
    ],
)
def test_max_labels_linear(max_labels, expected):
    colorbar = cm.LinearColormap(["red"] * 10, vmin=0, vmax=9, max_labels=max_labels)
    try:
        colorbar.render()
    except AssertionError:  # rendering outside parent Figure raises error
        pass
    assert colorbar.tick_labels == expected


@pytest.mark.parametrize(
    "max_labels,expected",
    [
        (10, [0.0, "", 2.0, "", 4.0, "", 6.0, "", 8.0, "", 10.0, ""]),
        (5, [0.0, "", "", 3.0, "", "", 6.0, "", "", 9.0, "", ""]),
        (3, [0.0, "", "", "", 4.0, "", "", "", 8.0, "", "", ""]),
    ],
)
def test_max_labels_step(max_labels, expected):
    colorbar = cm.StepColormap(
        ["red", "blue"] * 5,
        vmin=0,
        vmax=10,
        max_labels=max_labels,
    )
    try:
        colorbar.render()
    except AssertionError:  # rendering outside parent Figure raises error
        pass
    assert colorbar.tick_labels == expected



"""
Utilities
-------

Utility module for Folium helper functions.

"""

import base64
import json
import math
import os
import re
import struct
import typing
import zlib
from typing import Any, Callable, List, Optional, Sequence, Tuple, Union

from jinja2 import Environment, PackageLoader

try:
    import numpy as np
except ImportError:
    np = None  # type: ignore

if typing.TYPE_CHECKING:
    from branca.colormap import ColorMap


rootpath: str = os.path.abspath(os.path.dirname(__file__))


TypeParseSize = Union[int, float, str, Tuple[float, str]]


def get_templates() -> Environment:
    """Get Jinja templates."""
    return Environment(loader=PackageLoader("branca", "templates"))


def legend_scaler(
    legend_values: Sequence[float],
    max_labels: int = 10,
) -> List[Union[float, str]]:
    """
    Downsamples the number of legend values so that there isn't a collision
    of text on the legend colorbar (within reason). The colorbar seems to
    support ~10 entries as a maximum.

    """
    legend_ticks: List[Union[float, str]]
    if len(legend_values) < max_labels:
        legend_ticks = list(legend_values)
    else:
        spacer = int(math.ceil(len(legend_values) / max_labels))
        legend_ticks = []
        for i in legend_values[::spacer]:
            legend_ticks += [i]
            legend_ticks += [""] * (spacer - 1)
    return legend_ticks


def linear_gradient(hexList: List[str], nColors: int) -> List[str]:
    """
    Given a list of hexcode values, will return a list of length
    nColors where the colors are linearly interpolated between the
    (r, g, b) tuples that are given.
    """

    def _scale(start, finish, length, i):
        """
        Return the value correct value of a number that is in between start
        and finish, for use in a loop of length *length*.

        """
        base = 16

        fraction = float(i) / (length - 1)
        raynge = int(finish, base) - int(start, base)
        thex = hex(int(int(start, base) + fraction * raynge)).split("x")[-1]
        if len(thex) != 2:
            thex = "0" + thex
        return thex

    allColors: List[str] = []
    # Separate (R, G, B) pairs.
    for start, end in zip(hexList[:-1], hexList[1:]):
        # Linearly interpolate between pair of hex ###### values and
        # add to list.
        nInterpolate = 765
        for index in range(nInterpolate):
            r = _scale(start[1:3], end[1:3], nInterpolate, index)
            g = _scale(start[3:5], end[3:5], nInterpolate, index)
            b = _scale(start[5:7], end[5:7], nInterpolate, index)
            allColors.append("".join(["#", r, g, b]))

    # Pick only nColors colors from the total list.
    result: List[str] = []
    for counter in range(nColors):
        fraction = float(counter) / (nColors - 1)
        index = int(fraction * (len(allColors) - 1))
        result.append(allColors[index])
    return result


def color_brewer(color_code: str, n: int = 6) -> List[str]:
    """
    Generate a colorbrewer color scheme of length 'len', type 'scheme.
    Live examples can be seen at http://colorbrewer2.org/

    """
    maximum_n = 253
    minimum_n = 3

    if not isinstance(n, int):
        raise TypeError("n has to be an int, not a %s" % type(n))

    # Raise an error if the n requested is greater than the maximum.
    if n > maximum_n:
        raise ValueError(
            "The maximum number of colors in a"
            " ColorBrewer sequential color series is 253",
        )
    if n < minimum_n:
        raise ValueError(
            "The minimum number of colors in a"
            " ColorBrewer sequential color series is 3",
        )

    if not isinstance(color_code, str):
        raise ValueError(f"color should be a string, not a {type(color_code)}.")
    if color_code[-2:] == "_r":
        base_code = color_code[:-2]
        core_color_code = base_code + "_" + str(n).zfill(2)
        color_reverse = True
    else:
        base_code = color_code
        core_color_code = base_code + "_" + str(n).zfill(2)
        color_reverse = False

    with open(os.path.join(rootpath, "_schemes.json")) as f:
        schemes = json.loads(f.read())

    with open(os.path.join(rootpath, "scheme_info.json")) as f:
        scheme_info = json.loads(f.read())

    with open(os.path.join(rootpath, "scheme_base_codes.json")) as f:
        core_schemes = json.loads(f.read())["codes"]

    if base_code not in core_schemes:
        raise ValueError(base_code + " is not a valid ColorBrewer code")

    explicit_scheme = True
    if schemes.get(core_color_code) is None:
        explicit_scheme = False

    # Only if n is greater than the scheme length do we interpolate values.
    if not explicit_scheme:
        # Check to make sure that it is not a qualitative scheme.
        if scheme_info[base_code] == "Qualitative":
            matching_quals = []
            for key in schemes:
                if base_code + "_" in key:
                    matching_quals.append(int(key.split("_")[1]))

            raise ValueError(
                "Expanded color support is not available"
                " for Qualitative schemes; restrict the"
                " number of colors for the "
                + base_code
                + " code to between "
                + str(min(matching_quals))
                + " and "
                + str(max(matching_quals)),
            )
        else:
            longest_scheme_name = base_code
            longest_scheme_n = 0
            for sn_name in schemes.keys():
                if "_" not in sn_name:
                    continue
                if sn_name.split("_")[0] != base_code:
                    continue
                if int(sn_name.split("_")[1]) > longest_scheme_n:
                    longest_scheme_name = sn_name
                    longest_scheme_n = int(sn_name.split("_")[1])

            if not color_reverse:
                color_scheme = linear_gradient(schemes.get(longest_scheme_name), n)
            else:
                color_scheme = linear_gradient(
                    schemes.get(longest_scheme_name)[::-1],
                    n,
                )
    else:
        if not color_reverse:
            color_scheme = schemes.get(core_color_code, None)
        else:
            color_scheme = schemes.get(core_color_code, None)[::-1]
    return color_scheme


def image_to_url(
    image: Any,
    colormap: Union["ColorMap", Callable, None] = None,
    origin: str = "upper",
) -> str:
    """Infers the type of an image argument and transforms it into a URL.

    Parameters
    ----------
    image: string, file or array-like object
        * If string, it will be written directly in the output file.
        * If file, it's content will be converted as embedded in the
          output file.
        * If array-like, it will be converted to PNG base64 string and
          embedded in the output.
    origin : ['upper' | 'lower'], optional, default 'upper'
        Place the [0, 0] index of the array in the upper left or
        lower left corner of the axes.
    colormap : ColorMap or callable, used only for `mono` image.
        Function of the form [x -> (r,g,b)] or [x -> (r,g,b,a)]
        for transforming a mono image into RGB.
        It must output iterables of length 3 or 4, with values between
        0. and 1.  Hint : you can use colormaps from `matplotlib.cm`.
    """
    if hasattr(image, "read"):
        # We got an image file.
        if hasattr(image, "name"):
            # We try to get the image format from the file name.
            fileformat = image.name.lower().split(".")[-1]
        else:
            fileformat = "png"
        url = "data:image/{};base64,{}".format(
            fileformat,
            base64.b64encode(image.read()).decode("utf-8"),
        )
    elif (not (isinstance(image, str) or isinstance(image, bytes))) and hasattr(
        image,
        "__iter__",
    ):
        # We got an array-like object.
        png = write_png(image, origin=origin, colormap=colormap)
        url = "data:image/png;base64," + base64.b64encode(png).decode("utf-8")
    else:
        # We got an URL.
        url = json.loads(json.dumps(image))

    return url.replace("\n", " ")


def write_png(
    data: Any,
    origin: str = "upper",
    colormap: Union["ColorMap", Callable, None] = None,
) -> bytes:
    """
    Transform an array of data into a PNG string.
    This can be written to disk using binary I/O, or encoded using base64
    for an inline PNG like this:

    >>> png_str = write_png(array)
    >>> "data:image/png;base64," + png_str.encode("base64")

    Inspired from
    http://stackoverflow.com/questions/902761/saving-a-numpy-array-as-an-image

    Parameters
    ----------
    data: numpy array or equivalent list-like object.
         Must be NxM (mono), NxMx3 (RGB) or NxMx4 (RGBA)
    origin : ['upper' | 'lower'], optional, default 'upper'
        Place the [0,0] index of the array in the upper left or lower left
        corner of the axes.
    colormap : ColorMap subclass or callable, optional
        Only needed to transform mono images into RGB. You have three options:
        - use a subclass of `ColorMap` like `LinearColorMap`
        - use a colormap from `matplotlib.cm`
        - use a custom function of the form [x -> (r,g,b)] or [x -> (r,g,b,a)].
          It must output iterables of length 3 or 4 with values between 0 and 1.

    Returns
    -------
    PNG formatted byte string
    """
    from branca.colormap import ColorMap

    if np is None:
        raise ImportError("The NumPy package is required" " for this functionality")

    if isinstance(colormap, ColorMap):
        colormap_callable = colormap.rgba_floats_tuple
    elif callable(colormap):
        colormap_callable = colormap
    else:
        colormap_callable = lambda x: (x, x, x, 1)  # noqa E731

    array = np.atleast_3d(data)
    height, width, nblayers = array.shape

    if nblayers not in [1, 3, 4]:
        raise ValueError("Data must be NxM (mono), " "NxMx3 (RGB), or NxMx4 (RGBA)")
    assert array.shape == (height, width, nblayers)

    if nblayers == 1:
        array = np.array(list(map(colormap_callable, array.ravel())))
        nblayers = array.shape[1]
        if nblayers not in [3, 4]:
            raise ValueError(
                "colormap must provide colors of" "length 3 (RGB) or 4 (RGBA)",
            )
        array = array.reshape((height, width, nblayers))
    assert array.shape == (height, width, nblayers)

    if nblayers == 3:
        array = np.concatenate((array, np.ones((height, width, 1))), axis=2)
        nblayers = 4
    assert array.shape == (height, width, nblayers)
    assert nblayers == 4

    # Normalize to uint8 if it isn't already.
    if array.dtype != "uint8":
        with np.errstate(divide="ignore", invalid="ignore"):
            array = array * 255.0 / array.max(axis=(0, 1)).reshape((1, 1, 4))
            array[~np.isfinite(array)] = 0
        array = array.astype("uint8")

    # Eventually flip the image.
    if origin == "lower":
        array = array[::-1, :, :]

    # Transform the array to bytes.
    raw_data = b"".join([b"\x00" + array[i, :, :].tobytes() for i in range(height)])

    def png_pack(png_tag, data):
        chunk_head = png_tag + data
        return (
            struct.pack("!I", len(data))
            + chunk_head
            + struct.pack("!I", 0xFFFFFFFF & zlib.crc32(chunk_head))
        )

    return b"".join(
        [
            b"\x89PNG\r\n\x1a\n",
            png_pack(b"IHDR", struct.pack("!2I5B", width, height, 8, 6, 0, 0, 0)),
            png_pack(b"IDAT", zlib.compress(raw_data, 9)),
            png_pack(b"IEND", b""),
        ],
    )


def _camelify(out: str) -> str:
    return (
        (
            "".join(
                [
                    (
                        "_" + x.lower()
                        if i < len(out) - 1 and x.isupper() and out[i + 1].islower()
                        else (
                            x.lower() + "_"
                            if i < len(out) - 1 and x.islower() and out[i + 1].isupper()
                            else x.lower()
                        )
                    )
                    for i, x in enumerate(list(out))
                ],
            )
        )
        .lstrip("_")
        .replace("__", "_")
    )


def _parse_size(value: TypeParseSize) -> Tuple[float, str]:
    if isinstance(value, (int, float)):
        return float(value), "px"
    elif isinstance(value, str):
        # match digits or a point, possibly followed by a space,
        # followed by a unit: either 1 to 5 letters or a percent sign
        match = re.fullmatch(r"([\d.]+)\s?(\w{1,5}|%)", value.strip())
        if match:
            return float(match.group(1)), match.group(2)
        else:
            raise ValueError(
                f"Cannot parse {value!r}, it should be a number followed by a unit.",
            )
    elif (
        isinstance(value, tuple)
        and isinstance(value[0], (int, float))
        and isinstance(value[1], str)
    ):
        # value had been already parsed
        return (float(value[0]), value[1])
    else:
        raise TypeError(
            f"Cannot parse {value!r}, it should be a number or a string containing a number and a unit.",
        )


def _locations_mirror(x):
    """Mirrors the points in a list-of-list-of-...-of-list-of-points.
    For example:
    >>> _locations_mirror([[[1, 2], [3, 4]], [5, 6], [7, 8]])
    [[[2, 1], [4, 3]], [6, 5], [8, 7]]

    """
    if hasattr(x, "__iter__"):
        if hasattr(x[0], "__iter__"):
            return list(map(_locations_mirror, x))
        else:
            return list(x[::-1])
    else:
        return x


def _locations_tolist(x):
    """Transforms recursively a list of iterables into a list of list."""
    if hasattr(x, "__iter__"):
        return list(map(_locations_tolist, x))
    else:
        return x


def none_min(x: Optional[float], y: Optional[float]) -> Optional[float]:
    if x is None:
        return y
    elif y is None:
        return x
    else:
        return min(x, y)


def none_max(x: Optional[float], y: Optional[float]) -> Optional[float]:
    if x is None:
        return y
    elif y is None:
        return x
    else:
        return max(x, y)


def iter_points(x: Union[List, Tuple]) -> list:
    """Iterates over a list representing a feature, and returns a list of points,
    whatever the shape of the array (Point, MultiPolyline, etc).
    """
    if isinstance(x, (list, tuple)):
        if len(x):
            if isinstance(x[0], (list, tuple)):
                out = []
                for y in x:
                    out += iter_points(y)
                return out
            else:
                return [x]
        else:
            return []
    else:
        raise ValueError(f"List/tuple type expected. Got {x!r}.")



import json
import os
from pathlib import Path

import pytest

import branca.utilities as ut
from branca.colormap import LinearColormap

rootpath = Path(os.path.dirname(os.path.abspath(__file__))) / ".." / "branca"
color_brewer_minimum_n = 3
color_brewer_maximum_n = 253  # Why this limitation @ branca/utilities.py:108 ?


# Loads schemes and their meta-data
with open(rootpath / "_schemes.json") as f:
    schemes = json.loads(f.read())
with open(rootpath / "scheme_info.json") as f:
    scheme_info = json.loads(f.read())
with open(rootpath / "scheme_base_codes.json") as f:
    core_schemes = json.loads(f.read())["codes"]


def test_color_brewer_base():
    scheme = ut.color_brewer("YlGnBu", 9)
    assert scheme == [
        "#ffffd9",
        "#edf8b1",
        "#c7e9b4",
        "#7fcdbb",
        "#41b6c4",
        "#1d91c0",
        "#225ea8",
        "#253494",
        "#081d58",
    ]


def test_color_brewer_reverse():
    scheme = ut.color_brewer("YlGnBu")
    scheme_r = ut.color_brewer("YlGnBu_r")
    assert scheme[::-1] == scheme_r


def test_color_brewer_extendability():
    """
    The non-qualitative schemes should be extendable.

    :see https://github.com/python-visualization/branca/issues/104
    :see https://github.com/python-visualization/branca/issues/114

    Moreover, the following error was not reported via issues:
    * TypeError in the linear_gradient function when trying to extend any scheme.
    Indeed, in color_brewer, the key searched in the scheme database was not found,
    thus, it was passing `None` instead of a real scheme vector to linear_gradient.
    """
    for sname in core_schemes:
        for n in range(color_brewer_minimum_n, color_brewer_maximum_n + 1):
            try:
                scheme = ut.color_brewer(sname, n=n)
            except Exception as e:
                if scheme_info[sname] == "Qualitative" and isinstance(e, ValueError):
                    continue
                raise

            assert len(scheme) == n

            # When we try to extend a scheme,
            # the reverse is not always the exact reverse vector of the original one.
            # Thus, we do not test this property!
            _ = ut.color_brewer(sname + "_r", n=n)


def test_color_avoid_unexpected_error():
    """
    We had unexpected errors by providing some scheme name with unexpected value of `n`.
    This function tests them.

    Identified errors which was not reported via issues:
    * The scheme 'viridis' was not in the base_codes JSON;
    * Multiple scheme hadn't any metadata in scheme_info JSON;
    * When a `n` value provided to `color_scheme` was a float,
    it tried to select an unknown scheme without raising the right Exception type.
    """

    # Verify that every scheme has is present in base codes
    scheme_names = set()
    for sname in schemes.keys():
        scheme_names.add(sname.split("_")[0])
    assert scheme_names == set(core_schemes)

    # Verify that every scheme has a metadata
    assert scheme_names == set(scheme_info.keys())

    # Verify that every scheme can be generated in color_brewer using exotic value of `n`.
    # Note that big but valid values are generated by test_color_brewer_extendability.
    for sname in scheme_names:
        for n in (
            [-10]
            + list(range(-1, color_brewer_minimum_n))
            + list(range(color_brewer_maximum_n + 1, color_brewer_maximum_n + 10))
        ):
            with pytest.raises(ValueError):
                ut.color_brewer(sname, n)
        for n in [str(color_brewer_minimum_n), float(color_brewer_minimum_n), "abc"]:
            with pytest.raises(TypeError):
                ut.color_brewer(sname, n)


@pytest.mark.parametrize(
    "value,result",
    [
        (1, (1.0, "px")),
        ("1 px", (1.0, "px")),
        ("80 % ", (80.0, "%")),
        ("100%   ", (100.0, "%")),
        ("3 vw", (3.0, "vw")),
        ("3.14 rem", (3.14, "rem")),
        ((1, "px"), (1.0, "px")),
        ((80.0, "%"), (80.0, "%")),
    ],
)
def test_parse_size(value, result):
    assert ut._parse_size(value) == result


@pytest.mark.parametrize(
    "value",
    [
        "what?",
        "1.21 jigawatts",
        ut._parse_size,
        (1.21, 4.9),
    ],
)
def test_parse_size_exceptions(value):
    with pytest.raises((ValueError, TypeError)):
        ut._parse_size(value)


def test_write_png_mono():
    mono_image = [
        [0.24309289, 0.75997446, 0.02971671, 0.52830537],
        [0.62339252, 0.65369358, 0.41545387, 0.03307279],
    ]

    mono_png = b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x04\x00\x00\x00\x02\x08\x06\x00\x00\x00\x7f\xa8}c\x00\x00\x00)IDATx\xdac\x08\x0c\x0c\xfc\x0f\x02\x9c\x9c\x9c\xff7n\xdc\xf8\x9f\xe1\xe2\xc5\x8b\xffo\xdf\xbe\xfd\xbf\xbb\xbb\xfb?77\xf7\x7f\x00f\x87\x14\xdd\x0c\r;\xc0\x00\x00\x00\x00IEND\xaeB`\x82"  # noqa E501
    assert ut.write_png(mono_image) == mono_png

    colormap = LinearColormap(colors=["red", "yellow", "green"])
    color_png = b"\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x04\x00\x00\x00\x02\x08\x06\x00\x00\x00\x7f\xa8}c\x00\x00\x00)IDATx\xdac\xf8_\xcf\xf0\xbf\xea\x10\xc3\xff\xff\xfc\x0c\xff?\xfcg\xf8\xcfp\xe0\x19\xc3\xff\r\xf7\x80\x02\xb7\x80X\x90\xe1?\x00N\xca\x13\xcd\xfb\xad\r\xb8\x00\x00\x00\x00IEND\xaeB`\x82"  # noqa E501
    assert ut.write_png(mono_image, colormap=colormap) == color_png


def test_write_png_rgb():
    image_rgb = [
        [
            [0.8952778565195247, 0.6196806506704735, 0.2696137085302287],
            [0.3940794236804127, 0.9432178293916365, 0.16500617914697335],
            [0.5566755388192485, 0.10469673377265687, 0.27346260130585975],
            [0.2029951628162342, 0.5357152681832641, 0.13692921080346832],
        ],
        [
            [0.5186482474007286, 0.8625240370164696, 0.6965561989987038],
            [0.04425586727957387, 0.45448042432657076, 0.8552600511205423],
            [0.696453974598333, 0.7508742900711168, 0.9646572952994652],
            [0.7471809029502141, 0.3218907599994758, 0.789193070740859],
        ],
    ]
    png = b'\x89PNG\r\n\x1a\n\x00\x00\x00\rIHDR\x00\x00\x00\x04\x00\x00\x00\x02\x08\x06\x00\x00\x00\x7f\xa8}c\x00\x00\x00-IDATx\xda\x01"\x00\xdd\xff\x00\xff\xa7G\xffp\xff+\xff\x9e\x1cH\xff9\x90$\xff\x00\x93\xe9\xb8\xff\x0cz\xe2\xff\xc6\xca\xff\xff\xd4W\xd0\xffYw\x15\x95\xcf\xb9@D\x00\x00\x00\x00IEND\xaeB`\x82'  # noqa E501
    assert ut.write_png(image_rgb) == png
